/*
 * var_set_word to take care of set of words as compared to set of bits
 * Created by Malay K. Ganai (malay@ece.utexas.edu) on 1999/07/17 
 *
 */
#include "hash.h"
#include "util.h"
#include "var_set.h"
#include "var_set_word.h"

#define VAR_SET_PRIME_KEY_SIZE 4

static unsigned prime_array[] = {
 83, 311, 599, 997
};


var_set_word_t *var_set_word_new(size)
int size;
{
  var_set_word_t *result = ALLOC(var_set_word_t,1);

  result->n_elts = size;
  result->n_words = size / VAR_SET_WORD_SIZE + ((size % VAR_SET_WORD_SIZE == 0) ? 0 : 1);
  result->data = ALLOC(unsigned int, result->n_words * VAR_SET_WORD_SIZE);
  (void) var_set_word_clear(result);
  return result;
}

var_set_word_t *var_set_word_copy(set)
var_set_word_t *set;
{
  int i;
  var_set_word_t *result = ALLOC(var_set_word_t, 1);

  *result = *set;
  result->data = ALLOC(unsigned int, result->n_words * VAR_SET_WORD_SIZE);
  for (i = 0; i < result->n_words * VAR_SET_WORD_SIZE; i++)
    result->data[i] = set->data[i];
  return result;
}

void var_set_word_assign(result, set)
var_set_word_t *result;
var_set_word_t *set;
{
  int i;

  assert(result->n_elts == set->n_elts);
  for (i = 0; i < result->n_words * VAR_SET_WORD_SIZE; i++)
    result->data[i] = set->data[i];
}

void var_set_word_assign_elt(set, index, word)
var_set_word_t *set;
int index;
unsigned word;
{
  int i;
  unsigned *value;
  assert(index >= 0 && index < set->n_elts);
  
  for(i=0;i<VAR_SET_WORD_SIZE;i++){
    value = &(set->data[i * set->n_words + (index / VAR_SET_WORD_SIZE)]);
    if ( word & ( 1<<i)){  
      *value = *value | (1 << (index % VAR_SET_WORD_SIZE));
    } else {
      *value = *value & ~(1 << (index % VAR_SET_WORD_SIZE));
    }
  }
}


void var_set_word_free(set)
var_set_word_t *set;
{
  FREE(set->data);
  FREE(set);
}


void var_set_word_clear(set)
var_set_word_t *set;
{
  int i;

  for (i = 0; i < set->n_words * VAR_SET_WORD_SIZE; i++)
    set->data[i] = 0;
}



/* to compress the var_set_word data to 32 bit using non-trivial hashing */

void var_set_word_compress(compress_data, set, init_val)
unsigned compress_data[];
var_set_word_t *set;
unsigned init_val;
{

#ifndef NEW_HASH
  int i=0,j,k;

  unsigned  result[VAR_SET_PRIME_KEY_SIZE];
  
  for(k=0;k<VAR_SET_WORD_SIZE;k++){
    compress_data[k] = 0;
  
    for(j=0; j < VAR_SET_PRIME_KEY_SIZE ; j++){
      result[j] = 0;
    }
    for(; i< (k+1)* set->n_words ; i++){
      for(j=0; j < VAR_SET_PRIME_KEY_SIZE ; j++){
        result[j] = result[j] *  prime_array[j] + set->data[i];
      }
    }
    for(j=0;j<VAR_SET_PRIME_KEY_SIZE;j++){
      compress_data[k] += result[j];
    }
  }
#else
  unsigned *words = set->data;
  register int k;
  for(k=0;k<VAR_SET_WORD_SIZE;k++){
    compress_data[k] = hash(words,set->n_words, init_val);
    words += set->n_words;
  }
#endif  
}

