/**CHeaderFile*****************************************************************

  FileName                   [siva.h]

  PackageName                [siva_core]

  Synopsis                   [header  for siva]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modified: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]
******************************************************************************/

#ifndef SIVA_H
#define SIVA_H

#define SIVA_STATE_DBG 0
#define SIVA_DEBUG 0


#include "st.h"
#include "array.h"
#include "list.h"
#include "var_set.h"
#include "var_set_word.h"
#include "util.h"
#include "bck.h"
#include "cmd.h"
#include "vm.h"
#include <assert.h>
#include "fgrp_Defines.hh"
#include <sched.h>
#include <time.h>
#include <unistd.h>

#define SIVA_NODE BCK

/* symbolic solvers */
#define SIVA_ATPG 0
#define SIVA_SC   1

#define SIVA_DEFAULT_ERROR_CHECK_LEVEL 0
#define SIVA_DEFAULT_VERBOSITY_LEVEL 1
#define SIVA_DEFAULT_MAX_TIME_LIMIT 1000000
#define SIVA_DEFAULT_SYMBOLIC_SIM_DEPTH 100
#define SIVA_DEFAULT_ATPG_BACK_TRACK_LIMIT 100000
#define SIVA_DEFAULT_ATPG_NUM_TRIAL 2

#define SIVA_DEFAULT_ATPG_UNFOLD_DEPTH_UPPER 5 
#define SIVA_DEFAULT_ATPG_UNFOLD_DEPTH_LOWER 0
#define SIVA_DEFAULT_PARTITION_OVERLAP_THRESHOLD 0.001
#define SIVA_ATPG_UNFOLD_DEPTH_MAX 1000

#define SIVA_NUM_WORD(x) (((x-1)>>5)+1)


#define  SIVA_DEFAULT_ATPG_CONFLCIT_LIMIT 200000
#define  SIVA_DEFAULT_ATPG_DEDUCTION_LEVEL 0 
#define  SIVA_DEFAULT_ATPG_LEARNING_LEVEL 0
#define  SIVA_DEFAULT_PRE_PROCESS_NUM_TRIAL 1



#define SIVA_DEFAULT_LATCH_ACTIVITY_THRESHOLD 1.0

#define SIVA_UNREACHABLE_ET_MAX 20
#define SIVA_DEFAULT_SIM_DEPTH 16
#define SIVA_SIM_N_WORDS 128
#define SIVA_WORD_SIZE 32

#define SIVA_PRINT_FREQ 0x3f

#ifndef LIGHT_HOUSE
#define SIVA_STATE_DEPTH_BITS 27
#else
#define SIVA_STATE_DEPTH_BITS 14
#define SIVA_LIGHT_HOUSE_INDEX_BITS 13
#endif

#define SIVA_STATE_FLAG_BIT 1

/* entity types */
#define SIVA_NODE_ENTITY 0
#define SIVA_SUBCKT_ENTITY 1



#define SIVA_SEED(x) ((x) & (1<<26) -1)

#ifdef GPROF
#define SIVA_ASSERT(cond,msg)
#define SIVA_WARN1(cond,f_Str,msg)
#define SIVA_ASSERT(cond,msg) 
#define SIVA_ASSERT1(cond,f_str,arg))
#define SIVA_ASSERT2(cond,f_str,arg1,arg2)
#define SIVA_WARN(cond,f_str) 
#define SIVA_WARN1(cond,f_str,arg)                              
#define SIVA_VERB1_ARG1(f_str) 
#define SIVA_VERB1_ARG2(f_str,arg1) 
#define SIVA_VERB1_ARG3(f_str,arg1,arg2)
#define SIVA_VERB1_ARG4(f_str,arg1,arg2,arg3)
#define SIVA_VERB1_ARG5(f_str,arg1,arg2,arg3,arg4) 
#define SIVA_VERB2_ARG1(f_str) 
#define SIVA_VERB2_ARG2(f_str,arg1) 
#define SIVA_VERB2_ARG3(f_str,arg1,arg2) 
#define SIVA_VERB2_ARG4(f_str,arg1,arg2,arg3) 
#define SIVA_VERB3_ARG1(f_str) 
#define SIVA_VERB3_ARG2(f_str,arg1) 
#define SIVA_VERB3_ARG3(f_str,arg1,arg2)
#define SIVA_VERB3_ARG4(f_str,arg1,arg2,arg3) 
#define SIVA_VERB1(x) 
#define SIVA_VERB2(x) 
#define SIVA_VERB3(x) 
#define SIVA_DBG1(f_str) 
#define SIVA_DBG2(f_str,arg1) 
#define SIVA_DBG3(f_str,arg1,arg2) 
#define SIVA_ERR_CHK1(x) 
#define SIVA_ERR_CHK2(x) 
#define SIVA_ERR_CHK3(x) 

#else

#define SIVA_ASSERT(cond,msg) ((cond)==0)?printf("%s\n",msg):0;\
                              assert(cond)
#define SIVA_ASSERT1(cond,f_str,arg) ((cond)==0)?printf(f_str,arg):0;\
                              assert(cond)
#define SIVA_ASSERT2(cond,f_str,arg1,arg2)\
                             ((cond)==0)?printf(f_str,arg1,arg2):0;\
                              assert(cond)
#define SIVA_WARN(cond,f_str) ((cond)==0)?printf(f_str):0
#define SIVA_WARN1(cond,f_str,arg) ((cond)==0)?printf(f_str,arg):0

#define SIVA_VERB0_ARG1(f_str) ((siva_verbosity_level >= 0)?\
                                fprintf(siva_output_file,f_str):0)
#define SIVA_VERB0_ARG2(f_str,arg1) ((siva_verbosity_level >= 0)?\
                                fprintf(siva_output_file,f_str,arg1):0)
#define SIVA_VERB0_ARG3(f_str,arg1,arg2) ((siva_verbosity_level >= 0)?\
                                fprintf(siva_output_file,f_str,arg1,arg2):0)
#define SIVA_VERB0_ARG4(f_str,arg1,arg2,arg3) ((siva_verbosity_level >= 0)?\
                                fprintf(siva_output_file,f_str,arg1,arg2,arg3):0)
#define SIVA_VERB0_ARG5(f_str,arg1,arg2,arg3,arg4) ((siva_verbosity_level >= 0)?\
                                fprintf(siva_output_file,f_str,arg1,arg2,arg3,arg4):0)

#define SIVA_VERB0_ARG6(f_str,arg1,arg2,arg3,arg4,arg5) ((siva_verbosity_level >= 0)?\
                                fprintf(siva_output_file,f_str,arg1,arg2,arg3,arg4,arg5):0)
                              
#define SIVA_VERB1_ARG1(f_str) ((siva_verbosity_level >= 1)?\
                                fprintf(siva_output_file,f_str):0)
#define SIVA_VERB1_ARG2(f_str,arg1) ((siva_verbosity_level >= 1)?\
                                fprintf(siva_output_file,f_str,arg1):0)
#define SIVA_VERB1_ARG3(f_str,arg1,arg2) ((siva_verbosity_level >= 1)?\
                                fprintf(siva_output_file,f_str,arg1,arg2):0)
#define SIVA_VERB1_ARG4(f_str,arg1,arg2,arg3) ((siva_verbosity_level >= 1)?\
                                fprintf(siva_output_file,f_str,arg1,arg2,arg3):0)
#define SIVA_VERB1_ARG5(f_str,arg1,arg2,arg3,arg4) ((siva_verbosity_level >= 1)?\
                                fprintf(siva_output_file,f_str,arg1,arg2,arg3,arg4):0)

#define SIVA_VERB1_ARG6(f_str,arg1,arg2,arg3,arg4,arg5) ((siva_verbosity_level >= 1)?\
                                fprintf(siva_output_file,f_str,arg1,arg2,arg3,arg4,arg5):0)

#define SIVA_VERB1_ARG7(f_str,arg1,arg2,arg3,arg4,arg5,arg6) ((siva_verbosity_level >= 1)?\
                                fprintf(siva_output_file,f_str,arg1,arg2,arg3,arg4,arg5,arg6):0)


#define SIVA_VERB2_ARG1(f_str) ((siva_verbosity_level >= 2)?\
                                fprintf(siva_output_file,f_str):0)
#define SIVA_VERB2_ARG2(f_str,arg1) ((siva_verbosity_level >= 2)?\
                                fprintf(siva_output_file,f_str,arg1):0)
#define SIVA_VERB2_ARG3(f_str,arg1,arg2) ((siva_verbosity_level >= 2)?\
                                fprintf(siva_output_file,f_str,arg1,arg2):0)
#define SIVA_VERB2_ARG4(f_str,arg1,arg2,arg3) ((siva_verbosity_level >= 2)?\
                                fprintf(siva_output_file,f_str,arg1,arg2,arg3):0)
#define SIVA_VERB2_ARG5(f_str,arg1,arg2,arg3,arg4) ((siva_verbosity_level >= 2)?\
                                fprintf(siva_output_file,f_str,arg1,arg2,arg3,arg4):0)

#define SIVA_VERB2_ARG6(f_str,arg1,arg2,arg3,arg4,arg5) ((siva_verbosity_level >= 2)?\
                                fprintf(siva_output_file,f_str,arg1,arg2,arg3,arg4,arg5):0)


#define SIVA_VERB2_ARG7(f_str,arg1,arg2,arg3,arg4,arg5,arg6) ((siva_verbosity_level >= 2)?\
                                fprintf(siva_output_file,f_str,arg1,arg2,arg3,arg4,arg5,arg6):0)

#define SIVA_VERB3_ARG1(f_str) ((siva_verbosity_level >= 3)?\
                                fprintf(siva_output_file,f_str):0)
#define SIVA_VERB3_ARG2(f_str,arg1) ((siva_verbosity_level >= 3)?\
                                fprintf(siva_output_file,f_str,arg1):0)
#define SIVA_VERB3_ARG3(f_str,arg1,arg2) ((siva_verbosity_level >= 3)?\
                                fprintf(siva_output_file,f_str,arg1,arg2):0)
#define SIVA_VERB3_ARG4(f_str,arg1,arg2,arg3) ((siva_verbosity_level >= 3)?\
                                fprintf(siva_output_file,f_str,arg1,arg2,arg3):0)
#define SIVA_VERB3_ARG5(f_str,arg1,arg2,arg3,arg4) ((siva_verbosity_level >= 3)?\
                                fprintf(siva_output_file,f_str,arg1,arg2,arg3,arg4):0)

#define SIVA_VERB1(x) ((siva_verbosity_level>=1)? (x) : 0)
#define SIVA_VERB2(x) ((siva_verbosity_level>=2)? (x) : 0)
#define SIVA_VERB3(x) ((siva_verbosity_level>=3)? (x) : 0)

#define SIVA_ERR_CHK1(x) ((siva_error_check_level>=1)? (x) : 0)
#define SIVA_ERR_CHK2(x) ((siva_error_check_level>=2)? (x) : 0)
#define SIVA_ERR_CHK3(x) ((siva_error_check_level>=3)? (x) : 0)

#define SIVA_DBG1(f_str) ((SIVA_DEBUG)?\
                                fprintf(siva_output_file,f_str):0)
#define SIVA_DBG2(f_str,arg1) ((SIVA_DEBUG)?\
                                fprintf(siva_output_file,f_str,arg1):0)
#define SIVA_DBG3(f_str,arg1,arg2) ((SIVA_DEBUG)?\
                                fprintf(siva_output_file,f_str,arg1,arg2):0)

#endif

#define SIVA_IMPLY(x,y) (!(x) || (y))

/* complex node can be of the following type */

#define INPUT 0
#define LATCH 1
#define GATE 2
#define CONSTANT 3


/* companding method */
#define NORMAL 0
#define LOG 1
#define ALOG 2
#define SIVA_DEFAULT_MAX_NUM_ACTIVE_LATCH_GROUP 0



typedef struct siva_state SIVA_STATE;
typedef struct siva_state_aux SIVA_STATE_AUX;
typedef struct siva_table SIVA_TABLE;
typedef struct siva_complex_node SIVA_COMPLEX_NODE;
typedef struct siva_latch_node SIVA_LATCH_NODE;
typedef union  siva_key SIVA_KEY;
typedef struct siva_control_state SIVA_CONTROL_STATE;
typedef struct siva_hot_condition SIVA_HOT_CONDITION;
typedef struct siva_light_house SIVA_LIGHT_HOUSE;
typedef struct siva_module SIVA_MODULE;
typedef struct siva_module_entity SIVA_MODULE_ENTITY;
typedef struct siva_state_wrapper SIVA_STATE_WRAPPER;
typedef struct siva_iter SIVA_ITER;
typedef struct siva_et SIVA_ET;
typedef union  siva_conv SIVA_CONV;
typedef struct siva_heap SIVA_HEAP;
typedef struct siva_list SIVA_LIST;


/* function typedef */
typedef void (*STORE_WITNESS_SRV)
  (SIVA_STATE* state, var_set_t * vector, int rel_depth,
    int unfold_depth,void *para);

typedef void (*PRINT_WITNESS_SRV)  (SIVA_STATE* state,int unfold_depth);


struct siva_state
{
  unsigned state_value;
  struct siva_state *next; /* for hash chain */  
};


struct siva_state_aux
{
  unsigned state_value;
  unsigned prev_state_value;
  unsigned rel_depth;
  unsigned *state_data;
  struct siva_state_aux *next;
};


struct siva_table
{
  struct siva_complex_node **input;
  unsigned type           : 2;
  unsigned latch_input    : 1;
  unsigned target_zero    : 1;
  unsigned target_one     : 1;
  unsigned constraint_zero: 1; /* Future use */
  unsigned constraint_one : 1; /* Future use */
  unsigned out_phase      : 1;
  unsigned inv_support    : 1; /* Future use */
  unsigned num_input      : 9;
  unsigned num_minterm    : 12;
  unsigned debug_node     : 1;
  unsigned debug_target   : 1; /* Future use */
  /* dont-care = (entry[i][j] & (1<<l+1))
     bit-value = (entry[i][j] & (1<<l))
     will give the bit for the kth
     input in the ith minterm where  j = k/16, l = 2*(k%32) */
  unsigned **entry; 
};

struct siva_complex_node
{
  struct siva_table tbl;
  SIVA_NODE bck;
  array_t *light_house_array;
  char *name;
  unsigned level:    31;
  unsigned visited:    1;  
  struct siva_complex_node *next;
};



struct siva_latch_node
{
  SIVA_NODE latch;
  SIVA_NODE input;
};





union siva_key
{
  int int_key;
  float float_key;
  char *str_key;
  unsigned unsigned_key;
};

struct siva_control_state
{
  unsigned control_state_value;
  struct siva_control_state *next;
};

struct siva_hot_condition
{
  unsigned strong_zero   : 1;
  unsigned strong_one    : 1;
  unsigned weak_zero     : 1;
  unsigned weak_one      : 1;
  unsigned cold          : 1;
  unsigned weak_conflict : 1;
  
};

struct siva_light_house
{
  union {
    SIVA_NODE node;
    SIVA_ET *et;
  } target;
  SIVA_NODE node;
  unsigned priority : 20;
  unsigned is_target_et : 1;
};

struct siva_module
{
  char    *name;
  array_t *input_array;
  array_t *latch_array;
  array_t *output_array;
  st_table *entity_table;
  array_t *entity_array;
};

/* This structure is not efficient. Need to think about it */
struct siva_module_entity
{
  char     *name;
  unsigned type;
  union{
    struct {
      array_t  *input_array;
      unsigned input_array_dup : 1;
      unsigned out_phase       : 1;
      unsigned type            : 2;
      array_t *minterm_array;
    } node;
    struct {
      char    *formal_name;
      array_t *formal_array;
      array_t *actual_array;
    } subckt;
  }entity;
};

struct siva_state_wrapper
{
  SIVA_STATE *state;
  var_set_t *state_vector;
  int strategy;
};



struct siva_iter
{
  int array_index;
  int array_size;
  int ptr_array_index;
  int ptr_array_size;
  void* array;
  int index;
};

struct siva_et
{
  unsigned *data;
  unsigned *mask;
  union {
    struct siva_et *et;
    SIVA_NODE node;
  } target;
  unsigned state_value;
  unsigned distance     : 10;
  unsigned is_target_et :  1;
  unsigned priority     : 20;
  unsigned is_unreachable : 1;
  struct siva_et *next;
};

union siva_conv
{
  void *ptr;
  const void *const_ptr;
  unsigned level; 
  SIVA_NODE bck;
  unsigned value;
  array_t *array;
  unsigned *value_ptr;
  SIVA_COMPLEX_NODE *complex_node;
  union {
    unsigned char byte[4];
  } *byte_ptr;
  SIVA_NODE **bck_ptr_ptr;
  var_set_t *vector;
  SIVA_STATE *state;
  int int_value;
  float float_value;
  char *str;
  char *char_ptr;
  struct siva_et *et;
  FILE *file;
  struct siva_light_house *lh;
};

struct siva_heap
{
  union siva_key key;
  union siva_conv value;
};

struct siva_list
{
  void *element;
  struct siva_list* next;
};



/* mlh */
unsigned *siva_mlh_state_priority;
/* alh */
unsigned *siva_alh_state_priority;


/* state update */
unsigned siva_update_state_depth;

/* wit state */
unsigned *siva_wit_state_pos;

/* names */
char *siva_input_design_name;
array_t *siva_complex_primary_input_array;
array_t *siva_complex_latch_input_array;
array_t *siva_complex_latch_array;
array_t *siva_complex_primary_output_array;

array_t *siva_target_node_array;




/* input file read status */
int siva_init_ext_interface_done;
int siva_solve_mode;

/* global var */
int siva_little_endian;

int siva_build_pi_free_bck_network_done;
int siva_control_state_check_on;
int siva_alh_on;
int siva_et_on;
int siva_print_witness_on;

int siva_error_check_level;
int siva_solve_resume;
int siva_bfs_state_on;
int siva_control_sub_group_on;
float siva_partition_overlap_threshold;
/* rand seed */
unsigned siva_rand_seed;
unsigned siva_current_rand_seed;
int siva_sim_rand_seed;

/* debug variables */
array_t *siva_debug_node_array;
array_t *siva_debug_node_name_array;

/* latch nodes */
SIVA_LATCH_NODE *siva_latch_nodes;
unsigned char *siva_latch_inversion_bytes;

/* primary input nodes */
SIVA_NODE *siva_primary_inputs;

/* num variables */
int siva_num_zero_target;
int siva_num_one_target;
int siva_num_primary_input;
int siva_num_latch;

int siva_latch_word_size;
int siva_primary_input_word_size;

/* solve method */
unsigned siva_solve_method;

/* time variables */
long siva_max_time_limit;

/* sim variables */
unsigned siva_sim_depth;
unsigned siva_current_sim_depth;
unsigned siva_pre_process_num_trial;

/* symbolic sim variables */
unsigned siva_symbolic_sim_depth;
unsigned siva_symbolic_sim_lower_depth;
int siva_sym_check_unreachable;
/* symbolic solvers works on bck, hence need to track the
   actual network nodes
*/
SIVA_NODE siva_last_network_node; 
unsigned siva_target_enlarge_factor;

/* latch activity */
float siva_latch_activity_threshold;
unsigned siva_init_latch_activity_done;
int siva_companding_method;
int *siva_active_latch_order;

/* atpg */
unsigned siva_atpg_back_track_limit;
unsigned siva_atpg_conflict_limit;
unsigned siva_atpg_deduction_level;
unsigned siva_atpg_learning_level;
unsigned siva_atpg_num_trial;
unsigned siva_atpg_unfold_depth_lower;
unsigned siva_atpg_unfold_depth_upper;
array_t *siva_target_group_array;
unsigned siva_atpg_check_unreachable;

/*  control state */
unsigned siva_max_num_control_group;
unsigned siva_control_group;
unsigned siva_num_control_group;
array_t *siva_control_group_mask_array;
var_set_t *siva_control_group_active_set;
array_t *siva_control_state_init_hash_value_array;



/* state */
SIVA_STATE *siva_initial_state;
unsigned siva_state_init_hash_value;
array_t *siva_initial_state_vector_array;
var_set_t *siva_initial_state_vector;
var_set_word_t *siva_state_word_vector;
var_set_t *siva_state_vector;
var_set_t *siva_input_vector;
unsigned  *siva_input_vector_data;
unsigned *siva_last_stored_state_depths;
SIVA_STATE **siva_last_stored_states;

/* verbose */
int siva_verbosity_level;

/* time frame */
int siva_unfold_depth;

/* pi free target */
array_t *siva_enlarged_target_array;

/* aux fields */
unsigned siva_global_aux;
unsigned siva_sim_space_aux;
unsigned siva_sim_lc_space_aux;
unsigned siva_sim_rc_space_aux;
unsigned siva_bdd_aux;
unsigned siva_sim_value_aux;
unsigned siva_control_group_id_aux;
unsigned siva_latch_id_aux;
unsigned siva_pi_id_aux;
unsigned siva_bck_name_aux;
unsigned siva_goal_var_id_aux;
unsigned siva_lh_aux;
unsigned siva_latch_inertial_aux;
/* flags */
unsigned siva_latch_flag;
unsigned siva_latch_input_flag;
unsigned siva_target_zero_flag;
unsigned siva_target_one_flag;
unsigned siva_print_flag;
unsigned siva_global_flag;
unsigned siva_sim_space_flag;
unsigned siva_sim_value_flag;
unsigned siva_loop_latch_flag;
unsigned siva_primary_input_flag;
unsigned siva_mlh_one_flag;
unsigned siva_mlh_zero_flag;
unsigned siva_alh_one_flag;
unsigned siva_alh_zero_flag;
unsigned siva_coi_flag;
unsigned siva_enlarged_target_one_flag;
unsigned siva_enlarged_target_zero_flag;
unsigned siva_constant_one_flag;
unsigned siva_constant_zero_flag;
unsigned siva_bck_visited_flag;


/* files */
FILE *siva_bck_out_file;
FILE *siva_debug_node_file;
FILE *siva_target_file;
FILE *siva_input_file;
FILE *siva_output_file;
FILE *siva_pruned_target_file;
FILE *siva_map_file;
FILE *siva_input_bias_file;
FILE *siva_output_blif_file;
FILE *siva_mv_file;
FILE *siva_blif_file;


/* manual light house */
unsigned *siva_mlh_state_pos;

/* siva_exec */
EXTERN int siva_exec_write_bck(void);
EXTERN int siva_exec_solve(void);
EXTERN int siva_exec_read_blif(void);
EXTERN int siva_exec_pre_process(void);
EXTERN int siva_exec_solve_init(void);
EXTERN int siva_exec_solve_quit(void);

/* siva_io */
EXTERN int siva_write_bck_file(void);
EXTERN int siva_read_debug_node_file(void);
EXTERN int siva_read_target_file(void);
EXTERN void siva_print_complex_node_srv(SIVA_COMPLEX_NODE *node);
EXTERN void siva_print_fanin_cone(SIVA_NODE node);
EXTERN void siva_write_target_file(void);
EXTERN void siva_print_index_state_vector(var_set_t *state_vector);
EXTERN void siva_print_name_state_vector(var_set_t *state_vector);
EXTERN void siva_print_index_input_vector(var_set_t *input_vector);
EXTERN void siva_print_name_input_vector(var_set_t *input_vector);
EXTERN void siva_target_stat(void);
EXTERN void siva_network_stat(void);



/* siva_state */
EXTERN void siva_build_initial_state_vector(void);
EXTERN void siva_free_initial_state_vector(void);
EXTERN void siva_create_initial_state(void);
EXTERN void siva_init_state(void);
EXTERN SIVA_STATE *siva_alloc_state(void);
EXTERN void siva_remove_state_hash(SIVA_STATE *state);
EXTERN void siva_insert_state_hash(SIVA_STATE *state);
EXTERN SIVA_STATE *siva_find_state_hash(unsigned state_value);
EXTERN void siva_init_state_table(void);
EXTERN void siva_quit_state_table(void);
EXTERN unsigned siva_get_num_state(void);
EXTERN void siva_init_state_aux(void);
EXTERN void siva_quit_state_aux(void);
EXTERN void siva_upload_state_vector(SIVA_STATE *state,
                                     var_set_t *state_vector);
EXTERN void siva_upload_state_aux(SIVA_STATE_AUX *state_aux);
EXTERN void siva_insert_state_aux_hash(SIVA_STATE_AUX *state_aux);
EXTERN SIVA_STATE_AUX *siva_alloc_state_aux(void);
EXTERN void siva_state_table_stat(void);
EXTERN SIVA_STATE * siva_get_or_create_state(unsigned state_value,
                                             unsigned prev_state_value,
                                             int rel_depth,
                                             unsigned *state_data,
                                             int *is_created);
EXTERN int siva_get_bfs_state(SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr);
EXTERN int siva_get_seed_bfs_state(SIVA_STATE **state_p,
                                   var_set_t *seed_state_vector);
EXTERN void siva_put_bfs_state(unsigned priority_value, SIVA_STATE *state);
EXTERN void siva_init_bfs_state(void);
EXTERN void siva_quit_bfs_state(void);
EXTERN int siva_verify_state(SIVA_STATE *state);
EXTERN int siva_print_state_trace(SIVA_STATE *state);
EXTERN void siva_print_witness_trace(var_set_t **input_vectors,
                                     var_set_t **state_vectors,
                                     int offset,
                                     int unfold_depth);

EXTERN int siva_regenerate_state_vector(SIVA_STATE *state,
                                        var_set_t *state_vector);

EXTERN void siva_build_witness_state(int word_index, int bit_index,
                                     var_set_t *state_vector);
EXTERN void siva_build_witness_input(int word_index, int bit_index,
                                     var_set_t *input_vector);

/* siva_util */
EXTERN SIVA_LATCH_NODE *siva_alloc_latch_node(int size);
EXTERN char *siva_alloc_copy_string(char *string);
EXTERN char *siva_alloc_n_copy_string(char *string,int size);
EXTERN void siva_free_string(char *string);
EXTERN void siva_free_node(SIVA_NODE *nodes);
EXTERN SIVA_NODE *siva_alloc_node(int size);
EXTERN void siva_free_latch_node(SIVA_LATCH_NODE *latch_nodes);
EXTERN void siva_order_latch_nodes(void);
EXTERN void siva_fill_fanin_cone(array_t *node_array,
                                 SIVA_COMPLEX_NODE *complex_node);
EXTERN void siva_init_stack(void);
EXTERN void siva_quit_stack(void);
EXTERN void siva_stack_push(void *entry);
EXTERN void *siva_stack_pop(void);
EXTERN int siva_stack_empty(void);
EXTERN int siva_clear_aux_srv(SIVA_NODE in,void *aux);
EXTERN void siva_init_aux(void **aux_field);
EXTERN void siva_init_witness(void);
EXTERN void siva_quit_witness(void);
EXTERN void siva_alloc_input_vector(SIVA_STATE *state, int num_input);
/*
EXTERN void siva_store_seed(SIVA_STATE *state,unsigned s, unsigned index);
EXTERN int siva_is_input_seed(SIVA_STATE *state);
EXTERN void siva_get_seed(SIVA_STATE *state,unsigned *s, unsigned *index);
EXTERN void siva_get_input_vector(SIVA_STATE *state, int curr_rel_depth);
EXTERN void siva_extract_input_vector(unsigned s, unsigned index);
*/
EXTERN char *siva_strrchr(char *string, char c, int nth_last);



EXTERN int siva_fill_target_srv(SIVA_NODE node, void *para);
EXTERN void siva_fill_targets(SIVA_NODE *nodes);
EXTERN void siva_set_state_vector(var_set_t *vector, unsigned *data);
EXTERN void siva_set_input_vector(var_set_t *vector, unsigned *data);
EXTERN int siva_get_fanin_cone_size(SIVA_NODE *nodes, int  num_node);
EXTERN void  siva_print_hotness(SIVA_HOT_CONDITION *hot_conds);
EXTERN int siva_get_latch_id(SIVA_NODE latch);
EXTERN SIVA_NODE siva_get_latch_from_id(int id);
EXTERN SIVA_NODE siva_get_latch_input_from_id(int id);

/* siva_sim */
EXTERN void siva_init_sim(void);
EXTERN void siva_quit_sim(void);
EXTERN void siva_sim_stat(void);
EXTERN void siva_update_state_table(void);
EXTERN void siva_update_current_with_next_state(void);
EXTERN void siva_update_current_with_seed_state_array(array_t* state_wrapper_array,
                                                      int num_state);
EXTERN void siva_update_current_with_seed_state(var_set_t *state_vector);

EXTERN void siva_generate_input(void);
EXTERN int siva_evaluate_pre_process_srv(SIVA_NODE node, void *para);
EXTERN int siva_evaluate_latch_activity_srv(SIVA_NODE node, void *para);
EXTERN int siva_evaluate_solve_srv(SIVA_NODE node, void *para);
EXTERN void siva_verify_simulation(void);
EXTERN void siva_valuate_latch(var_set_t *state_vector);
EXTERN void siva_verify_witness(var_set_t *state_vector,array_t *input_vector_array, SIVA_NODE target, int target_value);

EXTERN void siva_init_latch_activity(void);
EXTERN void siva_quit_latch_activity(void);
EXTERN void siva_fill_active_latch_group(array_t *group_vector_array,
                                         int *latch_order);
EXTERN int siva_get_latch_inertia(int latch_id, int value);
EXTERN void siva_evaluate_latch_activity(void);
EXTERN int siva_is_latch_active(SIVA_NODE latch, int group);
EXTERN void siva_init_update_state_table(void);
EXTERN void siva_quit_update_state_table(void);
EXTERN int siva_evaluate_bit(SIVA_NODE node, void *para);
EXTERN void siva_valuate_primary_input(var_set_t *input_vector);
EXTERN void siva_fill_state_vector(var_set_t *state_vector);
EXTERN void siva_update_latch_with_input(void);
EXTERN void siva_evaluate_node(SIVA_NODE *nodes,
                               int *node_values,
                               int num_node,
                               var_set_t *input_vector,
                               var_set_t *state_vector);
EXTERN int  siva_evaluate_sym_node(SIVA_NODE sym_node,
                                   var_set_t **input_vectors,
                                   SIVA_NODE **input_node_vectors,
                                   int sym_depth);


EXTERN int siva_verify_next_state_vector(var_set_t *next_state_vector,
                                         var_set_t *input_vector,
                                         var_set_t *current_state_vector);
EXTERN void  siva_sim_next_state_vector(var_set_t *next_state_vector,
                                        var_set_t *input_vector,
                                        var_set_t *current_state_vector);


/* siva_bck */
EXTERN SIVA_NODE siva_bck_xor(SIVA_NODE p1, SIVA_NODE p2);
EXTERN SIVA_NODE siva_bck_and(SIVA_NODE p1, SIVA_NODE p2, int phase1, int phase2);
EXTERN void siva_build_bck(SIVA_COMPLEX_NODE *node);
EXTERN void siva_init_bck(void);
EXTERN void siva_quit_bck(void);
EXTERN SIVA_NODE siva_get_bck_zero(void);
EXTERN SIVA_NODE siva_get_bck_one(void);
EXTERN SIVA_NODE siva_bck_or(SIVA_NODE p1, SIVA_NODE p2, int phase1, int phase2);
EXTERN void siva_bck_init_quantify_unfold(void);
EXTERN void siva_bck_quit_quantify_unfold(void);
EXTERN void siva_bck_quantify(SIVA_NODE *dests,
                              SIVA_NODE *nodes,int num_node);
EXTERN void siva_bck_unfold(SIVA_NODE *dests, SIVA_NODE *nodes,int num_node);
EXTERN int siva_bck_merge_and_propogate(int const_value, SIVA_NODE in);



/* siva_solve */
EXTERN void siva_solve_target(void);

/* siva_pre_process */
EXTERN void siva_pre_process_target(void);

/* siva_witness */
EXTERN void siva_generate_witness(var_set_t *state_vector, int depth,
                                  SIVA_NODE target, array_t *input_vcetor_array);

/* siva_control_state */
EXTERN void siva_init_control_state(void);
EXTERN void siva_quit_control_state(void);
EXTERN void siva_control_state(void);
EXTERN void siva_control_state_stat(void);
EXTERN SIVA_CONTROL_STATE *siva_find_control_state_hash(unsigned control_state_value);
EXTERN void siva_insert_control_state_hash(SIVA_CONTROL_STATE *control_state);
EXTERN void siva_remove_control_state_hash(SIVA_CONTROL_STATE *control_state);
EXTERN SIVA_CONTROL_STATE *siva_alloc_control_state(void);
EXTERN int siva_control_state_filter_in(var_set_t *state_vector);
EXTERN int siva_get_seed_control_state(SIVA_STATE **state_p,
                                              var_set_t *state_vector);
EXTERN void siva_put_control_state(float priority_value,SIVA_STATE *state);
EXTERN int siva_get_control_state(SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr);
EXTERN int siva_is_latch_active(SIVA_NODE latch, int group);

EXTERN void siva_control_project(var_set_t *state_vector, int control_group);


/* siva_pi_free */
EXTERN int siva_build_pi_free_bck_network(void);

/* siva_atpg */
EXTERN void siva_init_atpg(void);
EXTERN void siva_quit_atpg(void);
EXTERN int siva_atpg_find_cond(SIVA_NODE out, int target_value,
                        SIVA_HOT_CONDITION *hot_conds);

/* siva_auto_light_house */
EXTERN void siva_init_alh(void);
EXTERN void siva_quit_alh(void);
EXTERN void siva_update_alh(void);
EXTERN void siva_put_alh_state(unsigned priority_value, SIVA_STATE *state);
EXTERN int siva_get_alh_state(SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr);
EXTERN int siva_get_seed_alh_state(SIVA_STATE **state_p,
                                   var_set_t *seed_state_vector);
EXTERN void siva_generate_alh(SIVA_ET *et);
EXTERN void siva_find_alh_state(array_t *state_wrapper_array, int num_state);
EXTERN int siva_is_alh_redundant(SIVA_LIGHT_HOUSE* alh);
/* siva_manual_light_house */
EXTERN void siva_init_mlh(void);
EXTERN void siva_quit_mlh(void);
EXTERN void siva_build_mlh(void);
EXTERN void siva_update_mlh(void);
EXTERN void siva_find_mlh_state(array_t *state_wrapper_array, int num_state);
EXTERN int siva_get_seed_mlh_state(SIVA_STATE **state_p,
                                   var_set_t *seed_state_vector);
EXTERN void siva_put_mlh_state(unsigned priority_value, SIVA_STATE *state);
EXTERN int siva_get_mlh_state(SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr);

/* siva_atpg_solver */
EXTERN void siva_init_atpg_solver(void);
EXTERN void siva_quit_atpg_solver(void);
EXTERN int siva_atpg_sym_solve(SIVA_NODE *goals, SIVA_NODE *targets,
                               int *target_values,
                               int num_goal,
                               array_t *nodes_array,
                               int sym_depth,
                               SIVA_STATE_WRAPPER *state_wrapper,
                               void *para,
                               STORE_WITNESS_SRV store_witness_srv,
                               PRINT_WITNESS_SRV print_witness_srv);

EXTERN int siva_atpg_solve(SIVA_NODE *goals, int *goal_values, int num_goal,
                           array_t *state_vector_array, int num_state,
                           void *para, STORE_WITNESS_SRV store_witness_srv,
                           PRINT_WITNESS_SRV print_wtiness_srv);

EXTERN int siva_atpg_find_input_sequence(var_set_t *target_state_vector,
                                         var_set_t *seed_state_vector,
                                         int rel_distance,
                                         var_set_t **input_vectors,
                                         var_set_t **state_vectors);
EXTERN int siva_atpg_enlarge(SIVA_NODE *goals, int *goal_values,
                              int num_goal,
                              SIVA_ET *ets,
                              int e_factor);
EXTERN void siva_store_target_witness_state(SIVA_STATE *seed_state,
                                            var_set_t *wit_state_vector,
                                            int rel_depth,
                                            int unfold_depth,
                                            void *para);
EXTERN int siva_atpg_is_satisfiable(SIVA_NODE *goals, int *goal_values,
                                   int num_goal);
EXTERN void siva_atpg_print_witness_srv(SIVA_STATE *seed_state,
                                        int unfold_depth);
EXTERN void siva_atpg_print_witness_srv1(SIVA_STATE *seed_state,
                                         int unfold_depth);


/* siva_remove_redundant */
EXTERN void siva_remove_latch(void);
EXTERN void siva_remove_functional_dependent_latch(void);
EXTERN void siva_remove_redundant(void);
EXTERN void siva_update_latch(void);



/* siva_read_mv */
EXTERN int siva_exec_mv2blif(void);

/* siva witness state */
EXTERN void siva_init_wit_state(void);
EXTERN void siva_quit_wit_state(void);
EXTERN int siva_get_seed_wit_state(SIVA_STATE **state_p,
                                    var_set_t *state_vector);
EXTERN int siva_get_wit_state(SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr);
EXTERN void siva_put_wit_state(unsigned priority_value, SIVA_STATE *state);

/* siva vector-generation under constraint */
EXTERN void siva_init_vec_gen(void);
EXTERN void siva_vec_gen(void);
EXTERN void siva_quit_vec_gen(void);
EXTERN int  siva_check_legal_state(int wi);
EXTERN void siva_check_valid_vector(var_set_t *input_vector, var_set_t *state_vector);
EXTERN void siva_check_input_validity(int wi);

/* siva_enlarged_target */
EXTERN void siva_init_et(void);
EXTERN void siva_quit_et(void);
EXTERN void siva_do_for_all_et(void (*service)(SIVA_ET *et));
EXTERN SIVA_ET* siva_get_first_et(SIVA_ITER *iter);
EXTERN SIVA_ET* siva_get_next_et(SIVA_ITER *iter);
EXTERN unsigned siva_get_num_et(void);
EXTERN SIVA_ET* siva_alloc_et(void);
EXTERN void siva_enlarged_target_state(void);
EXTERN void siva_et_solve(array_t *state_wrapper_array, int num_state);
EXTERN SIVA_NODE siva_get_et_target_node(SIVA_ET *et);
EXTERN SIVA_ET* siva_enlarge_et(SIVA_ET *et);
EXTERN SIVA_ET* siva_enlarge_target(SIVA_NODE target, int target_value);
EXTERN void siva_et_stat(void);
EXTERN void siva_backward_sim(void);
EXTERN int siva_check_et_reached(SIVA_STATE *state,
                                 var_set_t *state_vector);

/* siva_symbolic_sim */
EXTERN void siva_init_symbolic_sim(void);
EXTERN void siva_quit_symbolic_sim(void);
EXTERN void siva_symbolic_sim(void);
EXTERN int siva_symbolic_sim1(SIVA_NODE* targets, int *goal_values,
                              int num_target,
                              SIVA_STATE_WRAPPER *state_wrapper,
                              int conjunction_mode,
                              void *para,
                              STORE_WITNESS_SRV store_witness_srv,
                              PRINT_WITNESS_SRV print_witness_srv);
EXTERN void siva_gc_symbolic_sim(void);

/* siva_heap */
EXTERN int siva_heap_int_max_cmp(SIVA_HEAP heap1, SIVA_HEAP heap2);
EXTERN int siva_heap_int_min_cmp(SIVA_HEAP heap1, SIVA_HEAP heap2);
EXTERN int siva_heap_unsigned_max_cmp(SIVA_HEAP heap1, SIVA_HEAP heap2);
EXTERN int siva_heap_unsigned_min_cmp(SIVA_HEAP heap1, SIVA_HEAP heap2);
EXTERN int siva_heap_float_max_cmp(SIVA_HEAP heap1, SIVA_HEAP heap2);
EXTERN int siva_heap_float_min_cmp(SIVA_HEAP heap1, SIVA_HEAP heap2);
EXTERN int siva_heap_str_max_cmp(SIVA_HEAP heap1, SIVA_HEAP heap2);
EXTERN int siva_heap_str_min_cmp(SIVA_HEAP heap1, SIVA_HEAP heap2);
EXTERN void siva_heap_init(void);
EXTERN void  siva_heap_quit(void);
EXTERN void siva_do_heap(unsigned id, void (*service)(int pos, SIVA_KEY *key_p, SIVA_CONV *value_p));
EXTERN unsigned siva_alloc_heap(int (*cmp_srv)(SIVA_HEAP heap1, SIVA_HEAP heap2));
EXTERN void  siva_release_heap(unsigned heap_id);
EXTERN void siva_heap_insert(SIVA_KEY key, SIVA_CONV value, unsigned heap_id);
EXTERN int siva_heap_extract_top(SIVA_KEY *key_ptr, SIVA_CONV *value_ptr, unsigned heap_id);
EXTERN int siva_heap_peek_top(SIVA_KEY *key_ptr, SIVA_CONV *value_ptr, unsigned id);
EXTERN void siva_heap_replace(int i, SIVA_KEY key, SIVA_CONV value, unsigned id);
EXTERN int siva_heap_size(unsigned id);
EXTERN void siva_heap_print(unsigned id);

/* siva_complex_node */
EXTERN int siva_node_cmp(const void *node_p1,const void *node_p2);
EXTERN int siva_complex_node_level_cmp(const void *node_p1,const void *node_p2);
EXTERN int siva_complex_node_bck_cmp(void *node_p1,void *node_p2);
EXTERN void siva_do_for_all_complex_node(void (*service)(SIVA_COMPLEX_NODE *node));
EXTERN void siva_init_complex_node(void);
EXTERN void siva_quit_complex_node(void);
EXTERN void siva_set_target_srv(SIVA_COMPLEX_NODE *node);
EXTERN void siva_clear_target_srv(SIVA_COMPLEX_NODE *node);
EXTERN void siva_levelize_complex_node_srv(SIVA_COMPLEX_NODE *node);
EXTERN SIVA_NODE siva_get_node(SIVA_COMPLEX_NODE *x);
EXTERN void siva_set_node(SIVA_COMPLEX_NODE *x,SIVA_NODE y);
EXTERN int siva_is_target_zero_set(SIVA_COMPLEX_NODE *x);
EXTERN int siva_is_target_one_set(SIVA_COMPLEX_NODE *x);
EXTERN void siva_set_target_one(SIVA_COMPLEX_NODE *x);
EXTERN void siva_set_target_zero(SIVA_COMPLEX_NODE *x);
EXTERN void siva_clear_target_one(SIVA_COMPLEX_NODE *x);
EXTERN void siva_clear_target_zero(SIVA_COMPLEX_NODE *x);
EXTERN unsigned siva_get_num_complex_node(void);
EXTERN void siva_insert_complex_node_hash(SIVA_COMPLEX_NODE *node);
EXTERN void siva_remove_complex_node_hash(SIVA_COMPLEX_NODE *node);
EXTERN SIVA_COMPLEX_NODE *siva_find_complex_node_hash(char *string);
EXTERN void siva_init_complex_node_table(void);
EXTERN void siva_quit_complex_node_table(void);
EXTERN SIVA_COMPLEX_NODE *siva_alloc_complex_node(unsigned type);
EXTERN void siva_alloc_complex_node_input(SIVA_COMPLEX_NODE *node,int size);
EXTERN void siva_alloc_complex_node_entry(SIVA_COMPLEX_NODE *node,int num_minterm);
EXTERN int siva_is_complex_node_primary_input(SIVA_COMPLEX_NODE *node);
EXTERN void siva_free_complex_node_input(SIVA_COMPLEX_NODE *node);
EXTERN void siva_free_complex_node_entry(SIVA_COMPLEX_NODE *node);
EXTERN void siva_clear_visit_srv(SIVA_COMPLEX_NODE *node);
EXTERN int siva_target_count_srv(SIVA_NODE node, void *para);
EXTERN void siva_clear_debug_node(SIVA_COMPLEX_NODE *node);
EXTERN void siva_set_debug_node(SIVA_COMPLEX_NODE *node);
EXTERN int siva_is_debug_node(SIVA_COMPLEX_NODE *node);
EXTERN int siva_is_latch_input(SIVA_COMPLEX_NODE *x);
#endif /* SIVA_H */



