/**CFile**********************************************************************

  FileName                   [siva_bck.c]

  PackageName                [siva_core]

  Synopsis                   [supports interface to bck functions]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modified: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]
******************************************************************************/

#include "siva.h"

#define SIVA_BINATE 0x3
#define SIVA_UNATE_ONE 0x2
#define SIVA_UNATE_ZERO 0x1
#define SIVA_INVERT_UNATE(x) ((x)^0x3)

/* BCK_INVALID is set to UINT_MAX (= -1) */
#define SIVA_START_AUX ((void *)(-2))
#define SIVA_END_AUX ((void *)(-3))



static SIVA_NODE siva_bck_xnor(SIVA_NODE p1, SIVA_NODE p2);
static SIVA_NODE siva_bck_imply(SIVA_NODE p1, SIVA_NODE p2);
static SIVA_NODE siva_bck_zero;
static SIVA_NODE siva_bck_one;
static array_t *siva_pi_sup_array;
static array_t *siva_current_sup_array;
static array_t *siva_next_sup_array;
static unsigned siva_unate_aux;
static unsigned siva_cofactor_aux;
static unsigned siva_reachable_count_aux;
static unsigned siva_order_aux;
static unsigned siva_fanin_cone_flag;

static void siva_fanin_cone_unate(SIVA_NODE node, int unate);
static void siva_bck_rank_primary_input(SIVA_NODE node, int num_pi_sup);
static int siva_bck_cmp_pi_rank(const void *node_p1,const void *node_p2);
static void siva_bck_count_reachable_node(SIVA_NODE node);
static void siva_init_aux1(void **aux_field);
static SIVA_NODE siva_bck_quantify1(SIVA_NODE node);

/*--------------------------------------------------------------------------*/
void
siva_init_bck(void)
{
  bck_init();
  siva_coi_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_coi_flag);
  siva_bck_visited_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_bck_visited_flag);
  siva_sim_value_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_sim_value_flag);
  siva_sim_space_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_sim_space_flag);
  siva_latch_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_latch_flag);
  siva_latch_input_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_latch_input_flag);
  siva_primary_input_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_primary_input_flag);
  siva_target_zero_flag = bck_alloc_flag();  
  bck_reset_all_flag(siva_target_zero_flag);
  siva_target_one_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_target_one_flag);
  siva_print_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_print_flag);
  siva_loop_latch_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_loop_latch_flag);
  siva_alh_one_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_alh_one_flag);
  siva_alh_zero_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_alh_zero_flag); 
  siva_mlh_one_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_mlh_one_flag);
  siva_mlh_zero_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_mlh_zero_flag);
  siva_enlarged_target_one_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_enlarged_target_one_flag);
  siva_enlarged_target_zero_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_enlarged_target_zero_flag);  
  siva_constant_zero_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_constant_zero_flag);
  siva_constant_one_flag = bck_alloc_flag();
  bck_reset_all_flag(siva_constant_one_flag);  
  siva_latch_inertial_aux = bck_alloc_aux(siva_init_aux);
  siva_latch_id_aux = bck_alloc_aux(siva_init_aux);
  siva_pi_id_aux = bck_alloc_aux(siva_init_aux);
  /* Since two complex node can map to same bck node,
     this naming is used for inputs and latches for
     witness generation in verbose mode
  */
  siva_bck_name_aux = bck_alloc_aux(siva_init_aux);
  siva_lh_aux = bck_alloc_aux(siva_init_aux);
  siva_bck_zero = bck_create_zero();
  siva_bck_one = bck_create_one();
  siva_enlarged_target_array = array_alloc(SIVA_NODE,16);
  siva_num_zero_target = 0;
  siva_num_one_target  = 0;
  siva_num_latch = 0;
  siva_num_primary_input = 0;


}

/*--------------------------------------------------------------------------*/
void
siva_quit_bck(void)
{
  array_free(siva_enlarged_target_array);  
  bck_release_aux(siva_lh_aux,NULL);
  bck_release_aux(siva_bck_name_aux,NULL);
  bck_release_aux(siva_pi_id_aux,NULL);
  bck_release_aux(siva_latch_id_aux,NULL);
  bck_release_aux(siva_latch_inertial_aux,NULL);
 
  bck_release_flag(siva_constant_zero_flag);
  bck_release_flag(siva_constant_one_flag);  
  bck_release_flag(siva_enlarged_target_zero_flag);
  bck_release_flag(siva_enlarged_target_one_flag);
  bck_release_flag(siva_alh_one_flag);
  bck_release_flag(siva_alh_zero_flag);  
  bck_release_flag(siva_mlh_one_flag);
  bck_release_flag(siva_mlh_zero_flag);  
  bck_release_flag(siva_loop_latch_flag);
  bck_release_flag(siva_print_flag);
  bck_release_flag(siva_target_one_flag);
  bck_release_flag(siva_target_zero_flag);
  bck_release_flag(siva_latch_input_flag);
  bck_release_flag(siva_latch_flag);
  bck_release_flag(siva_sim_space_flag);
  bck_release_flag(siva_sim_value_flag);
  bck_release_flag(siva_coi_flag);
  bck_quit();
}


/*--------------------------------------------------------------------------*/
SIVA_NODE
siva_bck_and(SIVA_NODE p1, SIVA_NODE p2, int phase1, int phase2)
{

  if (!phase1) {
    p1 = bck_create_not(p1);
  }
  if (!phase2) {
    p2 = bck_create_not(p2);
  }
  return bck_create_and(p1,p2);

}

/*--------------------------------------------------------------------------*/
SIVA_NODE
siva_bck_or(SIVA_NODE p1, SIVA_NODE p2, int phase1, int phase2)
{

  if (phase1) {
    p1 = bck_create_not(p1);
  }
  if (phase2) {
    p2 = bck_create_not(p2);
  }
  return bck_create_not(bck_create_and(p1,p2));

}


/*--------------------------------------------------------------------------*/
SIVA_NODE
siva_bck_xor(SIVA_NODE p1, SIVA_NODE p2)
{
  return siva_bck_or(bck_create_and(p1,bck_create_not(p2)),
                     bck_create_and(bck_create_not(p1),p2),1,1);  
}


/*--------------------------------------------------------------------------*/
static SIVA_NODE
siva_bck_xnor(SIVA_NODE p1, SIVA_NODE p2)
{
  return bck_create_not(siva_bck_xor(p1,p2));                        
}

/*--------------------------------------------------------------------------*/
static SIVA_NODE
siva_bck_imply(SIVA_NODE p1, SIVA_NODE p2)
{
  return siva_bck_or(bck_create_not(p1),p2,1,1);                        
}



/*--------------------------------------------------------------------------*/
int
siva_bck_merge_and_propogate(int const_value, SIVA_NODE in)
{
  if (const_value){
    return bck_set_input_constraint(in);
  }else {
    return bck_set_input_constraint(bck_create_not(in));
  }
}


/*--------------------------------------------------------------------------*/
SIVA_NODE
siva_get_bck_zero(void)
{
  return siva_bck_zero;
}

/*--------------------------------------------------------------------------*/
SIVA_NODE
siva_get_bck_one(void)
{
  return siva_bck_one;
}


/*--------------------------------------------------------------------------*/
void
siva_build_bck(SIVA_COMPLEX_NODE *node)
{
  
  if (!bck_is_null(siva_get_node(node))){
    return ;
  }

  if ( node->tbl.type == CONSTANT ) {
    if (node->tbl.out_phase){
      siva_set_node(node,bck_create_one());
    }else{
      siva_set_node(node,bck_create_zero());
    }
    return;
  } 


  if ( node->tbl.type == INPUT ) {
    siva_set_node(node,bck_create_var());
    siva_num_primary_input++;
    return;
  } 

  if  (node->tbl.type == LATCH ) {
    SIVA_NODE bck;
    bck = bck_create_var();
    siva_set_node(node,bck);
    siva_num_latch++;
    return;
  }

  if (!node->tbl.num_input){
    fprintf(siva_output_file,"Node: %s\n",node->name);
  }
  SIVA_ASSERT(node->tbl.num_input,"Node is used but not defined in blif?");
  {
    register int i;
    SIVA_COMPLEX_NODE *input_nodes[node->tbl.num_input];

    for(i=0;i<node->tbl.num_input;i++){
      input_nodes[i] = node->tbl.input[i];
    }
    
    qsort(input_nodes,node->tbl.num_input,sizeof(SIVA_COMPLEX_NODE *),
          siva_complex_node_level_cmp);

#if 1
    for(i=0;i<node->tbl.num_input;i++){
      siva_build_bck(input_nodes[i]);      
    }
#else
    for(i=node->tbl.num_input-1;i>=0;i--){
      siva_build_bck(input_nodes[i]);      
    }
#endif    
  }
  {
    SIVA_NODE bck_or;
    SIVA_NODE bck_and;
    register int i,j;

    /* create OR structure */
    bck_or = bck_create_zero();
    for(i=0;i<node->tbl.num_minterm;i++){
      bck_and = bck_create_one();
      for(j=0;j<node->tbl.num_input;j++){
        if ( node->tbl.entry[i][j/16] & (1<<(2*(j%16)+1) ))  {
          continue;
        }
        /* bit value is one */
        if ( node->tbl.entry[i][j/16] & (1<< (2*(j%16)) )) {      
          bck_and = siva_bck_and(bck_and,
                                 siva_get_node(node->tbl.input[j]),1,1);
        } else {
          bck_and = siva_bck_and(bck_and,
                                 siva_get_node(node->tbl.input[j]),1,0);
        }
      }
      bck_or = siva_bck_and(bck_and,bck_or,0,0);    
      if (node->tbl.out_phase) {
        bck_or = bck_create_not(bck_or);
      }
    }   
    siva_set_node(node,bck_or);    
    SIVA_VERB2_ARG3("node %s bck %d\n",node->name,
                    bck_get_index(bck_or));    
    return;
  }
}


/*--------------------------------------------------------------------------*/
void
siva_bck_init_quantify_unfold(void)
{
  int i;
  siva_pi_sup_array = array_alloc(SIVA_NODE,siva_num_primary_input);
  siva_current_sup_array = array_alloc(SIVA_NODE,siva_num_latch);
  siva_next_sup_array = array_alloc(SIVA_NODE,siva_num_latch);
  siva_reachable_count_aux = siva_cofactor_aux = bck_alloc_aux(siva_init_aux1);
  siva_order_aux = siva_unate_aux = bck_alloc_aux(siva_init_aux1);
  siva_fanin_cone_flag = bck_alloc_flag();

}


/*--------------------------------------------------------------------------*/
void
siva_bck_quit_quantify_unfold(void)
{
  int i;
  bck_release_flag(siva_fanin_cone_flag);
  bck_release_aux(siva_unate_aux,NULL);
  bck_release_aux(siva_cofactor_aux,NULL);
  array_free(siva_current_sup_array);
  array_free(siva_next_sup_array);    
  array_free(siva_pi_sup_array);
}


/*--------------------------------------------------------------------------*/
void
siva_bck_quantify(SIVA_NODE *dests, SIVA_NODE *srcs,
                  int num_dest)
{
  int k;
  SIVA_NODE nodes[num_dest];
  memcpy(nodes,srcs,sizeof(SIVA_NODE)*num_dest);  
  for(k=0;k<num_dest;k++){
    dests[k] = siva_bck_quantify1(nodes[k]);
  }
}

/*--------------------------------------------------------------------------*/
void
siva_bck_unfold(SIVA_NODE *dests, SIVA_NODE *srcs,
                int num_dest)
{
  int i;
  int num_sup;
  SIVA_NODE *in_v;
  SIVA_NODE *in_h;
  
  bck_reset_all_flag(siva_fanin_cone_flag);    
  for(i=0;i<num_dest;i++){
    bck_set_fanin_cone_flag(srcs[i],siva_fanin_cone_flag);
  }
  num_sup = 0;
  for(i=0;i<siva_num_latch;i++){
    if (!bck_is_flag_set(siva_latch_nodes[i].latch,siva_fanin_cone_flag)){
      continue;
    }
    array_insert(SIVA_NODE,siva_current_sup_array,num_sup,
                 bck_get_regular(siva_latch_nodes[i].latch));
    array_insert(SIVA_NODE,siva_next_sup_array,num_sup,
                 siva_latch_nodes[i].input);
    num_sup++;
  }
  {
    for(i=0;i<siva_num_primary_input;i++){     
      SIVA_ASSERT(!bck_is_flag_set(siva_primary_inputs[i],
                                   siva_fanin_cone_flag),
                  "Unfold of only PI-free nodes allowed");
    }
  }
  in_v = array_space(SIVA_NODE,siva_current_sup_array);
  in_h = array_space(SIVA_NODE,siva_next_sup_array);  
  bck_compose_vec(dests,srcs,num_dest,in_v,in_h,num_sup,NULL);
  
  for(i=0;i<num_dest;i++){
    bck_reset_fanin_cone_flag(srcs[i],siva_fanin_cone_flag);
  }

  SIVA_VERB2(bck_stat(siva_output_file));
  
}

/*--------------------------------------------------------------------------*/
static SIVA_NODE
siva_bck_quantify1(SIVA_NODE node)     
{
  int num_pi_sup;
  int i;
  SIVA_CONV conv;
  SIVA_NODE pos_cof;
  SIVA_NODE neg_cof;
  SIVA_NODE pi_free_node;
  
  bck_reset_all_flag(siva_fanin_cone_flag);  
  bck_set_fanin_cone_flag(node,siva_fanin_cone_flag);
  num_pi_sup=0;
  for(i=0;i<siva_num_primary_input;i++){
    if (!bck_is_flag_set(siva_primary_inputs[i],siva_fanin_cone_flag)){
      continue;
    }
    array_insert(SIVA_NODE,siva_pi_sup_array,num_pi_sup,
                 bck_get_regular(siva_primary_inputs[i]));
    num_pi_sup++;
  }
  if (num_pi_sup > 40 ){
    fprintf(siva_output_file,"Number of inputs support too large ..%d\n",
            num_pi_sup);
    bck_reset_fanin_cone_flag(node,siva_fanin_cone_flag);  
    return BCK_NULL;
  }

  if (siva_verbosity_level >= 4){
    bck_reset_fanin_cone_flag(node,siva_print_flag);
    siva_print_fanin_cone(node);
    bck_reset_fanin_cone_flag(node,siva_print_flag);
  }

  /* rank and order the primary inputs in the support of node*/
  /* use "wide eye" heuristic */
  /* this introduces element of ND in number of bck nodes */
#if 0
  siva_bck_rank_primary_input(node,num_pi_sup);
  array_sort1(siva_pi_sup_array,siva_bck_cmp_pi_rank,num_pi_sup);
#endif
  /* find the unate vars and mark them */
  bck_set_fanin_cone_aux(node,siva_unate_aux,SIVA_START_AUX);
  if (bck_is_inverted(node)){
    siva_fanin_cone_unate(node,SIVA_UNATE_ZERO);
  }else {
    siva_fanin_cone_unate(node,SIVA_UNATE_ONE);
  }
  
  /* quantify the node */
  pi_free_node = node;
  for(i=0;i<num_pi_sup;i++){
    SIVA_NODE pi = array_fetch(SIVA_NODE,siva_pi_sup_array,i);
    conv.ptr = bck_get_aux(pi,siva_unate_aux);
    if (conv.int_value == SIVA_UNATE_ONE){
      SIVA_VERB3_ARG2("Bck %d == unate one\n",pi);
    }
    if (conv.int_value == SIVA_UNATE_ZERO){
      SIVA_VERB3_ARG2("Bck %d == unate zero\n",pi);
    }
          
    if (conv.int_value == SIVA_BINATE || conv.int_value == SIVA_UNATE_ONE){
      pos_cof = bck_cofactor(pi_free_node,pi,siva_cofactor_aux);
    }else {
      pos_cof = siva_bck_zero;
    }
    if (conv.int_value == SIVA_BINATE || conv.int_value == SIVA_UNATE_ZERO){
      neg_cof = bck_cofactor(pi_free_node,bck_create_not(pi),
                             siva_cofactor_aux);
    }else {
      neg_cof = siva_bck_zero;
    }
    pi_free_node = siva_bck_or(pos_cof,neg_cof,1,1);
  }
  bck_set_fanin_cone_aux(node,siva_unate_aux,SIVA_END_AUX);

  bck_reset_fanin_cone_flag(node,siva_fanin_cone_flag);  
  return pi_free_node;
}

/*---------------------------------------------------------------------------*/
static void
siva_init_aux1(void **aux_field)
{
  (*aux_field) = SIVA_END_AUX;
}

/*--------------------------------------------------------------------------*/
static void
siva_fanin_cone_unate(SIVA_NODE node, int unate)
{
  SIVA_CONV conv;

  conv.ptr = bck_get_aux(node,siva_unate_aux);

  /* the node and its fanin cone are set to binate */
  /* or no contradiction yet */
  if (conv.int_value == SIVA_BINATE || conv.int_value == unate){
    return;
  }else if (conv.ptr == SIVA_START_AUX){
    conv.int_value = unate;
  }else {
    /*( conv.int_value != unate) */
    conv.int_value = SIVA_BINATE;
  }
  bck_set_aux(node,siva_unate_aux,conv.ptr);
  if (bck_is_var(node)){
    return;
  }
  if (bck_is_and(node)){
    SIVA_NODE left = bck_get_left_child(node);
    SIVA_NODE right = bck_get_right_child(node);
    if (conv.int_value == SIVA_BINATE){
      siva_fanin_cone_unate(left,conv.int_value);
      siva_fanin_cone_unate(right,conv.int_value);
    }else {      
      if (bck_is_inverted(left)){
        siva_fanin_cone_unate(left,SIVA_INVERT_UNATE(conv.int_value));
      }else {
        siva_fanin_cone_unate(left,conv.int_value);
      }
      if (bck_is_inverted(right)){
        siva_fanin_cone_unate(right,SIVA_INVERT_UNATE(conv.int_value));
      }else {
        siva_fanin_cone_unate(right,conv.int_value);
      }
    }
  }
  return;
}

/*--------------------------------------------------------------------------*/
static void
siva_bck_rank_primary_input(SIVA_NODE node, int num_pi_sup)
{
  int i;

  for(i=0;i<num_pi_sup;i++){
    SIVA_NODE pi = array_fetch(SIVA_NODE,siva_pi_sup_array,i);
    bck_set_fanin_cone_aux(node,siva_reachable_count_aux,SIVA_START_AUX);
    siva_bck_count_reachable_node(pi);
    bck_set_aux(pi,siva_order_aux,
                bck_get_aux(pi,siva_reachable_count_aux));
    bck_set_fanin_cone_aux(node,siva_reachable_count_aux,SIVA_END_AUX);
  }
}

/*---------------------------------------------------------------------------*/
static int
siva_bck_cmp_pi_rank(const void *node_p1,const void *node_p2)
{
  SIVA_CONV conv1;
  SIVA_CONV conv2;
  conv1.ptr = bck_get_aux(*(SIVA_NODE *)node_p1,siva_order_aux);
  conv2.ptr = bck_get_aux(*(SIVA_NODE *)node_p2,siva_order_aux);
  return (conv1.int_value - conv2.int_value);

}

/*--------------------------------------------------------------------------*/
static void
siva_bck_count_reachable_node(SIVA_NODE node)
{
  SIVA_NODE *outs;
  int num_out;
  int i;
  int num_reachable;
  SIVA_CONV conv;

  if(bck_get_aux(node,siva_reachable_count_aux) != SIVA_START_AUX) {
    return;
  }
  
  outs = bck_get_outputs(node);
  num_out = bck_get_num_outputs(node);

  for(i=0;i<num_out;i++){
    if (!bck_is_flag_set(outs[i],siva_fanin_cone_flag)) continue;
    siva_bck_count_reachable_node(outs[i]);
    conv.ptr = bck_get_aux(outs[i],siva_reachable_count_aux);
    num_reachable += conv.int_value;
  }
  conv.int_value = num_reachable;
  bck_set_aux(node,siva_reachable_count_aux,conv.ptr);
  return;
}



