/**CFile**********************************************************************

  FileName                   [siva_pi_free.c]

  PackageName                [siva_core]

  Synopsis                   [builds pi free network]

  Author                     [Malay K. Ganai]

  Comment                    [01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]


******************************************************************************/
#include "siva.h"

static void   siva_early_resolution(SIVA_NODE *pi_free_targets,
                                    SIVA_NODE *targets, int num_target);

/*--------------------------------------------------------------------------*/
int
siva_build_pi_free_bck_network(void)
{
  register int i,k;
  int num_target;
  SIVA_NODE *targets;
  SIVA_NODE *pi_free_targets;
  SIVA_NODE *unfold_targets;    
  SIVA_NODE p;
  int succeed = 1;
  if (siva_num_zero_target + siva_num_one_target == 0){
    SIVA_VERB1_ARG1("No targets to resolve\n");
    return;
  }

  if ((p = bck_next_node(siva_last_network_node)) != BCK_NULL) {
    bck_release_node(p);
  }

  /* initialize for targets */
  siva_target_stat();
  num_target = siva_num_zero_target+siva_num_one_target;
  targets = ALLOC(SIVA_NODE,num_target);  
  siva_fill_targets(targets);
  pi_free_targets = ALLOC(SIVA_NODE,num_target);
  unfold_targets = ALLOC(SIVA_NODE,num_target);

  
  siva_bck_init_quantify_unfold();
  for(i=0;i<num_target;i++){
    int obj = bck_is_flag_set(targets[i],siva_target_one_flag);
    SIVA_ASSERT((!bck_is_flag_set(targets[i],siva_target_one_flag) ||
                 !bck_is_flag_set(targets[i],siva_target_zero_flag)),
                "target can't be both one and zero");
    unfold_targets[i] = bck_get_regular(targets[i]);
    if (!obj){
      unfold_targets[i] = bck_create_not(unfold_targets[i]);
    }
  }


  
  SIVA_VERB1_ARG1("Enlarging targets ..\n");
  SIVA_VERB2_ARG1("Quantifying targets ..\n");  
  siva_bck_quantify(pi_free_targets,unfold_targets,num_target);
  for(k=0;k<num_target;k++){
    if (bck_is_null(pi_free_targets[k])) {
      SIVA_VERB1_ARG3("Failed to enlarge %d at depth %d\n",
                      bck_get_index(targets[k]),0);
      succeed = 0;
    }
  }
  if (!succeed) {
    SIVA_VERB1_ARG1("Aborting.\n");
    goto end;
  }       

  SIVA_VERB2_ARG1("Quantifying targets ..done\n");
  SIVA_VERB2(bck_stat(siva_output_file));

  siva_early_resolution(pi_free_targets,targets,num_target);  
  for(i=1;i<=siva_unfold_depth;i++){
    SIVA_VERB2_ARG2("Unfolding targets ..%d\n",i);
    siva_bck_unfold(unfold_targets,pi_free_targets,num_target);
    SIVA_VERB2_ARG1("Unfolding targets ..done\n");
    SIVA_VERB2_ARG1("Quantifying targets ..\n");
    siva_bck_quantify(pi_free_targets,unfold_targets,num_target);
    
    for(k=0;k<num_target;k++){
      if (bck_is_null(pi_free_targets[k])) {
        SIVA_VERB1_ARG3("Failed to enlarge %d at depth %d\n",
                        bck_get_index(targets[k]),i);
        succeed = 0;
      }
    }
    if (!succeed) {
      SIVA_VERB1_ARG1("Aborting. Try with lower depth\n");
      goto end;
    }       
    SIVA_VERB2_ARG1("Quantifying targets ..done\n");
    SIVA_VERB2_ARG1("Early resolution check ..\n");
    siva_early_resolution(pi_free_targets,targets,num_target);
    SIVA_VERB2_ARG1("Early resolution check ..done\n");
    
    if (siva_verbosity_level > 1){
      bck_stat(siva_output_file);
      siva_target_stat();
    }
  }

  for(i=0;i<num_target;i++){
    if (bck_is_constant(pi_free_targets[i])){
      SIVA_ASSERT(!bck_is_flag_set(targets[i],siva_target_one_flag) &&
                  !bck_is_flag_set(targets[i],siva_target_zero_flag),
                  "target should have been resolved\n");
      continue;
    }
    if (!bck_is_inverted(pi_free_targets[i])) {
      bck_set_flag(pi_free_targets[i],siva_enlarged_target_one_flag);
    }else {
      bck_set_flag(pi_free_targets[i],siva_enlarged_target_zero_flag);      
    }
    array_insert_last(SIVA_NODE,siva_enlarged_target_array,
                      targets[i]);
    array_insert_last(SIVA_NODE,siva_enlarged_target_array,
                      pi_free_targets[i]);
  }
  SIVA_VERB1_ARG1("Enlarging targets ..done\n");
  siva_last_network_node = bck_get_last_node();
 end:
  if ((p = bck_next_node(siva_last_network_node)) != BCK_NULL) {
    bck_release_node(p);
  }
  siva_bck_quit_quantify_unfold();
  FREE(targets);
  FREE(unfold_targets);
  FREE(pi_free_targets);

  SIVA_VERB1(siva_network_stat());
  SIVA_VERB1(siva_target_stat());
  
  return succeed;
}

/*--------------------------------------------------------------------------*/
static void  
siva_early_resolution(SIVA_NODE *pi_free_targets,
                      SIVA_NODE *targets, int num_target)
{
  int i;
  int values[num_target];

  siva_evaluate_node(pi_free_targets,values,num_target,
                     NIL(var_set_t),siva_initial_state_vector);
  for(i=0;i<num_target;i++){
    if (!bck_is_flag_set(targets[i],siva_target_one_flag) &&
        !bck_is_flag_set(targets[i],siva_target_zero_flag)){
      continue;
    }
    /* Caution: unreachability check is weak; use
       symbolic algorithm to prove unreachability
    */
    if (bck_is_zero(pi_free_targets[i])){
      fprintf(siva_output_file,"Target %d value %d is un-reachable\n",
              bck_get_index(targets[i]),
              bck_is_flag_set(targets[i],siva_target_one_flag));
      bck_reset_flag(targets[i],siva_target_one_flag);
      bck_reset_flag(targets[i],siva_target_zero_flag);
      continue;
    }
    if (values[i]){
      fprintf(siva_output_file,"Target %d value %d is reachable\n",
              bck_get_index(targets[i]),
              bck_is_flag_set(targets[i],siva_target_one_flag));
      bck_reset_flag(targets[i],siva_target_one_flag);
      bck_reset_flag(targets[i],siva_target_zero_flag);
      pi_free_targets[i] = bck_create_one();
      continue;
    }
  }
}
    
    
