/**CFile**********************************************************************

  FileName                   [siva_pre_process.c]

  PackageName                [siva_core]

  Synopsis                   [Solves targets using random simulation
  from initial state]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modified: 01/11/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]

******************************************************************************/

#include "siva.h"

/*
  --------------------------------------------------------------------------
  Nodes marked with siva_target_one_flag and siva_target_zero_flag
  are checked for reachability. If reachable, resets the flag.
  --------------------------------------------------------------------------
*/
void
siva_pre_process_target(void)
{
  register int i;
  int num_trial=0;
  long end_time;
  int print_freq;

  if (siva_num_one_target+siva_num_zero_target == 0) {
    return;
  }  
  end_time = util_cpu_time() + siva_max_time_limit;  
  siva_init_sim();
  SIVA_VERB1(siva_sim_stat());
  siva_init_witness();
  num_trial = 0;
  while(num_trial < siva_pre_process_num_trial){
    siva_update_current_with_seed_state(siva_initial_state_vector);
    siva_current_rand_seed = siva_rand_seed;
    print_freq = SIVA_PRINT_FREQ;
    for(i=0; i < siva_sim_depth; i++){
      if (--print_freq == 0) {
        SIVA_VERB1_ARG2("Simulation At depth %d\n",i);
        print_freq = SIVA_PRINT_FREQ;
      }
      siva_current_sim_depth = i;
      siva_generate_input();
      /* simulate */
      bck_do_in_topological_order(NULL,siva_evaluate_pre_process_srv);
      if (siva_num_one_target+siva_num_zero_target == 0) {
        break;
      }
      SIVA_ERR_CHK2(siva_verify_simulation());
      if (util_cpu_time() > end_time ) { 
        SIVA_VERB1_ARG1("Simulation time out\n");
        goto done;
      }
      siva_update_current_with_next_state();
    }
    num_trial++;
  }
 done:  
  siva_quit_witness();
  siva_quit_sim();

}

