/**CFile**********************************************************************

  FileName                   [siva_read_mv.c]

  PackageName                [siva_core]

  Synopsis                   [converts subset mv to blif; variables can
  have 2 values and ND is not supported]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modifed: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]
  
******************************************************************************/
#include "siva.h"

#define siva_alloc_module() (ALLOC(SIVA_MODULE,1))
#define siva_alloc_module_entity() (ALLOC(SIVA_MODULE_ENTITY,1))
#define siva_alloc_string(n) (ALLOC(char,n))


static char *siva_root_module_name;
static SIVA_MODULE *siva_root_module;
static st_table *siva_module_table;
static void siva_read_blif_mv(void);
static void siva_write_blif(void);
static void siva_write_blif_module(char *prefix,SIVA_MODULE *module);
static void siva_init_conversion(void);
static void siva_quit_conversion(void);

/*---------------------------------------------------------------------------*/
int
siva_exec_mv2blif(void)
{
  siva_init_conversion();
  siva_read_blif_mv();
  if (siva_blif_file){
    siva_write_blif();
  }
  siva_quit_conversion();
}

/*---------------------------------------------------------------------------*/
static void
siva_init_conversion(void)
{
  /* initialize */
  siva_root_module_name = NIL(char);
  siva_root_module = NIL(SIVA_MODULE);
  siva_module_table = st_init_table(strcmp,st_strhash);  
}

/*---------------------------------------------------------------------------*/
static void
siva_quit_conversion(void)
{
  st_generator *stgen1;
  st_generator *stgen2;
  SIVA_MODULE *module;
  SIVA_MODULE_ENTITY *entity;
  char *name;
  int i;
  
  st_foreach_item(siva_module_table,stgen1,(char **)&name,
                  (char **)&module){
    array_free(module->input_array);
    array_free(module->latch_array);
    array_free(module->output_array);
    siva_free_string(module->name);
    array_free(module->entity_array);
    st_foreach_item(module->entity_table,stgen2,(char **)&name,
                    (char **)&entity){
      siva_free_string(entity->name);
      if (entity->type == SIVA_NODE_ENTITY){
        if (entity->entity.node.type == INPUT ||
            entity->entity.node.type == CONSTANT){
          continue;
        }
        if (entity->entity.node.input_array &&
            !entity->entity.node.input_array_dup){
          for(i=0;i<array_n(entity->entity.node.input_array);i++){
            char *input_name =
              array_fetch(char *,entity->entity.node.input_array,i);
            siva_free_string(input_name);
          }
          array_free(entity->entity.node.input_array);
        }
        if (entity->entity.node.minterm_array){
          for(i=0;i<array_n(entity->entity.node.minterm_array);i++){
            var_set_t *minterm =
              array_fetch(var_set_t *,entity->entity.node.minterm_array,i);
            var_set_free(minterm);
          }
        }
      }else {
        siva_free_string(entity->entity.subckt.formal_name);
        for(i=0;i<array_n(entity->entity.subckt.formal_array);i++){
          char *formal_arg_name =
            array_fetch(char *,entity->entity.subckt.formal_array,i);
          siva_free_string(formal_arg_name);
        }
        for(i=0;i<array_n(entity->entity.subckt.actual_array);i++){
          char *actual_arg_name =
            array_fetch(char *,entity->entity.subckt.actual_array,i);
          siva_free_string(actual_arg_name);
        }
        array_free(entity->entity.subckt.formal_array);
        array_free(entity->entity.subckt.actual_array);
      }
    }
    st_free_table(module->entity_table);
  }
  st_free_table(siva_module_table);
                    
}

/*---------------------------------------------------------------------------*/
static void
siva_read_blif_mv(void)
{
  char buf[5000];
  char temp_buf[5000];
  char *token;
  char *first_token;
  char *list_of_tokens[500];
  int i,j;
  int num_token;
  char read_already=0;

  int input_parse_cont=0;
  int output_parse_cont=0;
  char first_time=0;

  SIVA_MODULE_ENTITY *node;
  SIVA_MODULE_ENTITY *latch_node;
  SIVA_MODULE_ENTITY *nodes[500];
  var_set_t *sets[500];
  SIVA_MODULE_ENTITY *subckt;
  
  SIVA_MODULE *current_module;
  char *first_module_name;
  char *node_name;
  char *latch_input_name;
  
  int num_output;
  int num_input;
  int num_argument;
  
    
  
  first_module_name = NIL(char);    
  SIVA_ASSERT(siva_mv_file,"input file not fopen-ed");

  while ( !feof(siva_mv_file) ){
    if (!read_already){
      fgets(buf,5000,siva_mv_file);
    }
    read_already=0;

    /* module name */
    if (( buf[0] == '.')&&( buf[1] == 'm')){
      token = strtok(buf," \n");
      token = strtok(NULL," \n") ;
      /* new module */
      current_module = siva_alloc_module();
      current_module->name = siva_alloc_copy_string(token);
      current_module->entity_table =
        st_init_table(strcmp,st_strhash);
      current_module->input_array = array_alloc(char *,4);
      current_module->latch_array = array_alloc(char *,4);
      current_module->output_array = array_alloc(char *,4);
      current_module->entity_array =
        array_alloc(SIVA_MODULE_ENTITY *,4);
      st_insert(siva_module_table,(char *)current_module->name,
                (char *)current_module);
      if (!first_module_name) {
        first_module_name = current_module->name;
      }
      continue;
    }
    
    /* root module */
    if (( buf[0] == '.')&&( buf[1] == 'r')&&(buf[2] == 'o')){
      SIVA_ASSERT1(!siva_root_module_name,
                    "%s is already defined root module\n",
                    siva_root_module_name);
      siva_root_module_name = current_module->name;
      continue;
    }

    /* inputs */

    if ((input_parse_cont)||(( buf[0] == '.')&&( buf[1] == 'i'))){
      strcpy(temp_buf,buf);
      first_time=1;
      while(1){
        if ( first_time ) {
           token = strtok(temp_buf,"\\ \t\n");
           first_time=0;
           if (!input_parse_cont) {
             continue;
           }
        }else {
           token = strtok(NULL,"\\ \t\n") ;
           if (token==NULL) {
             break;
           }
        }
        node = siva_alloc_module_entity();
        node->name = siva_alloc_copy_string(token);
        node->type = SIVA_NODE_ENTITY;
        node->entity.node.type = INPUT;
        st_insert(current_module->entity_table,(char *)node->name,
                  (char *)node);        
        array_insert_last(char *,current_module->input_array,
                          node->name);
        array_insert_last(SIVA_MODULE_ENTITY *,current_module->entity_array,
                          node);        
      }
      if (strrchr(buf,'\\')){
        input_parse_cont=1;
      }else {
        input_parse_cont=0;
      }
      continue;
    }

    /* outputs */

    if ((output_parse_cont)||(( buf[0] == '.')&&( buf[1] == 'o'))){
      strcpy(temp_buf,buf);
      first_time=1;
      while(1){
        if ( first_time ) {
           token = strtok(temp_buf,"\\ \t\n");
           first_time=0;
           if (!output_parse_cont) {
             continue;
           }
        }else {
           token = strtok(NULL,"\\ \t\n") ;
           if (token==NULL) {
             break;
           }
        }
        node_name = siva_alloc_copy_string(token);        
        array_insert_last(char *,
                          current_module->output_array,
                          node_name);
      }
      if (strrchr(buf,'\\')){
        output_parse_cont=1;
      }else {
        output_parse_cont=0;
      }
      continue;
    }

    /* latches */
    /* assuming that the latch has only input, output and reset parameter*/
    if (( buf[0] == '.')&&( buf[1] == 'l')){
      strcpy(temp_buf,buf);
      first_token = strtok(temp_buf," \t\n");
      num_token=0;
      while((token = strtok(NULL," \t\n")) != NULL ){
        list_of_tokens[num_token] = token;
        num_token++;
      }

      latch_node = siva_alloc_module_entity();
      latch_node->name = siva_alloc_copy_string(list_of_tokens[1]);
      latch_node->type = SIVA_NODE_ENTITY;
      latch_node->entity.node.input_array = array_alloc(char *,1);
      latch_node->entity.node.minterm_array = NIL(array_t);
      latch_node->entity.node.type = LATCH;
      latch_input_name = siva_alloc_copy_string(list_of_tokens[0]);
      array_insert(char *,latch_node->entity.node.input_array,0,
                   latch_input_name);
      array_insert_last(char *,
                        current_module->latch_array,
                        latch_node->name);
      st_insert(current_module->entity_table,(char *)latch_node->name,
                (char *)latch_node);
      array_insert_last(SIVA_MODULE_ENTITY *,current_module->entity_array,
                        latch_node);        
                     

      continue;
    }


    /* the names or the tables */

   if (( buf[0] == '.')&&( buf[1] == 'n')){
      strcpy(temp_buf,buf);
      first_token = strtok(temp_buf," \t\n");
      num_token=0;
      num_output = 0;
      while((token = strtok(NULL," \t\n")) != NULL ){
        if (strcmp("->",token) == 0){
          num_output = 0;
          continue;
        }
        list_of_tokens[num_token] = token;
        num_output++;
        num_token++;        
      }
      if (num_output == num_token){
        nodes[0] = siva_alloc_module_entity();
        nodes[0]->name = siva_alloc_copy_string(list_of_tokens[num_token-1]);
        nodes[0]->type = SIVA_NODE_ENTITY;
        nodes[0]->entity.node.type = GATE;
        nodes[0]->entity.node.input_array = NIL(array_t);
        nodes[0]->entity.node.minterm_array = NIL(array_t);
        
        nodes[0]->entity.node.out_phase = 1; /* my default */
        st_insert(current_module->entity_table,(char *)nodes[0]->name,
                  (char *)nodes[0]);
        array_insert_last(SIVA_MODULE_ENTITY *,current_module->entity_array,
                          nodes[0]);        

        num_input = num_token-1;
        num_output = 1;
      }else {
        num_input = num_token-num_output;
        for(i=0;i<num_output;i++){
          nodes[i] = siva_alloc_module_entity();
          nodes[i]->name =
            siva_alloc_copy_string(list_of_tokens[num_input+i]);
          nodes[i]->type = SIVA_NODE_ENTITY;
          nodes[i]->entity.node.type = GATE;
          nodes[i]->entity.node.out_phase = 1; /* my default */
          nodes[i]->entity.node.input_array = NIL(array_t);
          nodes[i]->entity.node.minterm_array = NIL(array_t);

          st_insert(current_module->entity_table,(char *)nodes[i]->name,
                    (char *)nodes[i]);
          array_insert_last(SIVA_MODULE_ENTITY *,current_module->entity_array,
                            nodes[i]);        

        }
      }
      if (num_token == 1){
        fgets(buf,5000,siva_mv_file);
        if (( buf[0] == '.')&&( buf[1] == 'd')){
          strcpy(temp_buf,buf);
          first_token = strtok(temp_buf," \t\n");      
          token = strtok(NULL," \t\n");
          nodes[0]->entity.node.out_phase = 1^atoi(token);
          fgets(buf,5000,siva_mv_file);
        }
        if (buf[0] == '1'){
          nodes[0]->entity.node.out_phase = 1;
        }else if (buf[0] == '0'){
          nodes[0]->entity.node.out_phase = 0;
        }else {
          read_already = 1;
        }
        nodes[0]->entity.node.type = CONSTANT;
        continue;
      }
      for(i=0;i<num_output;i++){
        nodes[i]->entity.node.minterm_array =
          array_alloc(var_set_t*,4);
      }
      nodes[0]->entity.node.input_array = array_alloc(char *,num_input);
      nodes[0]->entity.node.input_array_dup = 0;
      for(i=0;i<num_input;i++){
        char *fanin_name = siva_alloc_copy_string(list_of_tokens[i]); 
        array_insert(char*,nodes[0]->entity.node.input_array,i,fanin_name);
      }
      for(i=1;i<num_output;i++){
        nodes[i]->entity.node.input_array = nodes[0]->entity.node.input_array;
        nodes[i]->entity.node.input_array_dup = 1;
      }
      /* defaults */
      fgets(buf,5000,siva_mv_file);
      if (( buf[0] == '.')&&( buf[1] == 'd')){
        strcpy(temp_buf,buf);
        first_token = strtok(temp_buf," \t\n");      
        num_token=0;
        while((token = strtok(NULL," \t\n")) != NULL ){
          nodes[num_token]->entity.node.out_phase = 1^atoi(token);
          num_token++;        
        }
        fgets(buf,5000,siva_mv_file);
      }
      do {
        strcpy(temp_buf,buf);
        num_token=0;
        list_of_tokens[num_token++] = strtok(temp_buf," \t\n");
        while((token = strtok(NULL," \t\n")) != NULL ){
          list_of_tokens[num_token++] = token;
        }
        sets[0] = var_set_new(num_input * 2);
        /* assert(all clear) */
        for(i=0;i<num_input;i++){
          if (list_of_tokens[i][0] == '1'){
            var_set_set_elt(sets[0],2*i);
          }else if (list_of_tokens[i][0] == '-'){
            var_set_set_elt(sets[0],2*i+1);
          }else {
            /* if (list_of_tokens[i][0] == '0'){ */
          }
        }
        for(i=1;i<num_output;i++){
          sets[i] = var_set_copy(sets[0]);          
        }
        for(i=0;i<num_output;i++){
          if (list_of_tokens[num_input+i][0] == '=') {
            char *input_construct  = list_of_tokens[num_input+i]+1;
            int k;
            for(k=0;k<num_input;k++){
              char *input = array_fetch(char *,
                                        nodes[0]->entity.node.input_array,k);
              if (strcmp(input_construct,input) == 0){
                break;
              }
            }
            SIVA_ASSERT1((k<num_input),"input construct %s not found\n",
                        input_construct);
            if (nodes[i]->entity.node.out_phase){
              var_set_set_elt(sets[i],2*k);
              var_set_clear_elt(sets[i],2*k+1);
            }else {
              var_set_clear_elt(sets[i],2*k);
              var_set_clear_elt(sets[i],2*k+1);
            }
          }else{
            if ((list_of_tokens[num_input+i][0] == '0' &&
                 nodes[i]->entity.node.out_phase == 1) ||
                (list_of_tokens[num_input+i][0] == '1' &&
                 nodes[i]->entity.node.out_phase == 0)){           
              var_set_free(sets[i]);
              sets[i] = NIL(var_set_t);
            }
          }
        }
        for(i=0;i<num_output;i++){
          if (!sets[i]) continue;
          array_insert_last(var_set_t*,
                            nodes[i]->entity.node.minterm_array,
                            sets[i]);
        }
          
        fgets(buf,5000,siva_mv_file);
        if (( buf[0] == '.') || (buf[0] == '#')){
          read_already = 1;
          break;
        }
      } while (1);
   }
   
   if (( buf[0] == '.')&&( buf[1] == 's')){
      strcpy(temp_buf,buf);
      first_token = strtok(temp_buf," =\t\n");
      num_token=0;
      while((token = strtok(NULL," =\t\n")) != NULL ){
        list_of_tokens[num_token] = token;
        num_token++;        
      }
      num_argument = (num_token - 2) >> 1;
      subckt = siva_alloc_module_entity();
      subckt->name = siva_alloc_copy_string(list_of_tokens[1]);
      subckt->type = SIVA_SUBCKT_ENTITY;
      subckt->entity.subckt.formal_name =
        siva_alloc_copy_string(list_of_tokens[0]);
      subckt->entity.subckt.formal_array = array_alloc(char *,num_argument);
      subckt->entity.subckt.actual_array = array_alloc(char *,num_argument);
      for(i=0;i<num_argument;i++){
        char *formal_arg = siva_alloc_copy_string(list_of_tokens[i*2+2]);
        char *actual_arg = siva_alloc_copy_string(list_of_tokens[i*2+3]);
        array_insert(char *,subckt->entity.subckt.formal_array,
                     i,formal_arg);
        array_insert(char *,subckt->entity.subckt.actual_array,
                     i,actual_arg);
      }
      st_insert(current_module->entity_table,(char *)subckt->name,
                (char *)subckt);
      array_insert_last(SIVA_MODULE_ENTITY*,current_module->entity_array,
                        subckt);        

   }
  }

  if (siva_root_module_name == NIL(char)){
    siva_root_module_name = first_module_name;
  }

}


/*---------------------------------------------------------------------------*/
static void
siva_write_blif(void)
{
  int i;
  char *prefix;
  st_lookup(siva_module_table,(char *)siva_root_module_name,
            (char **)&siva_root_module);
  fprintf(siva_blif_file,"# Generated by mv2blif on %s\n","04/26/27");
  fprintf(siva_blif_file,".model %s\n",siva_root_module_name);
#if 1
  prefix = siva_alloc_string(strlen(siva_root_module_name)+2);
  strcat(prefix,siva_root_module_name);
  strcat(prefix,".");
#else
  prefix = NIL(char);
#endif
  for(i=0;i<array_n(siva_root_module->input_array);i++){
    char *input_name = array_fetch(char *,siva_root_module->input_array,i);
    fprintf(siva_blif_file,".inputs %s%s\n",prefix,
            input_name);
  }
  for(i=0;i<array_n(siva_root_module->output_array);i++){
    char *output_name = array_fetch(char *,siva_root_module->output_array,i);
    fprintf(siva_blif_file,".outputs %s%s\n",prefix,
            output_name);
  }
  siva_write_blif_module(prefix,siva_root_module);
#if 1
  siva_free_string(prefix);
#endif
  fprintf(siva_blif_file,".end\n");
}

/*---------------------------------------------------------------------------*/
static void
siva_write_blif_module(char *prefix, SIVA_MODULE *module)
{
  int i;
  int k;
  st_generator *stgen;
  char *name;
  SIVA_MODULE_ENTITY *entity;
#if 1
  st_foreach_item(module->entity_table,stgen,(char **)&name,
                  (char **)&entity){
#else
  for(k=0;k<array_n(module->entity_array);k++){
    entity = array_fetch(SIVA_MODULE_ENTITY*,module->entity_array,k);
#endif
    if (entity->type == SIVA_NODE_ENTITY){
      if (entity->entity.node.type == LATCH){
        char *latch_input_name =
          array_fetch(char *,entity->entity.node.input_array,0);
        fprintf(siva_blif_file,".latch %s%s %s%s %d\n",
                prefix,latch_input_name,prefix,entity->name,
                entity->entity.node.out_phase);
      }else  if (entity->entity.node.type == CONSTANT){
        fprintf(siva_blif_file,".names %s%s\n%d\n",
                prefix,entity->name,entity->entity.node.out_phase);
      }else if (entity->entity.node.type == GATE){
        fprintf(siva_blif_file,".names ");
        for(i=0;i<array_n(entity->entity.node.input_array);i++){
          char *input_name =
            array_fetch(char *,entity->entity.node.input_array,i);
          fprintf(siva_blif_file,"%s%s ",prefix,input_name);
        }
        fprintf(siva_blif_file,"%s%s\n",prefix,entity->name);
        for(i=0;i<array_n(entity->entity.node.minterm_array);i++){
          var_set_t *minterm =
            array_fetch(var_set_t *,entity->entity.node.minterm_array,i);
          int j;
          for(j=0;j<array_n(entity->entity.node.input_array);j++){
            if (var_set_get_elt(minterm,2*j+1)){
              fprintf(siva_blif_file,"-");
            }else {
              if (var_set_get_elt(minterm,2*j)){
                fprintf(siva_blif_file,"1");
              }else {
                fprintf(siva_blif_file,"0");
              }
            }
          }
          fprintf(siva_blif_file," %d\n",entity->entity.node.out_phase);
        }
      }
    }else {
      char *hier_name;
      int i;
      SIVA_MODULE *sub_module;
      hier_name = siva_alloc_string(strlen(prefix)+strlen(entity->name)+2);
      strcpy(hier_name,prefix);
      strcat(hier_name,entity->name);
      strcat(hier_name,".");
      st_lookup(siva_module_table,
                (char *)entity->entity.subckt.formal_name,
                (char **)&sub_module);
      /* write out the interfaces */
      fprintf(siva_blif_file,".C Interfaces between %s and module %s\n",
              hier_name,sub_module->name);
      for(i=0;i<array_n(entity->entity.subckt.formal_array);i++){
        char *formal_arg_name =
          array_fetch(char *,entity->entity.subckt.formal_array,i);
        char *actual_arg_name =
          array_fetch(char *,entity->entity.subckt.actual_array,i);
        SIVA_MODULE_ENTITY *sub_entity;
        st_lookup(sub_module->entity_table,(char *)formal_arg_name,
                  (char **)&sub_entity);
        if (sub_entity->entity.node.type == INPUT){
          fprintf(siva_blif_file,".names %s%s %s%s\n",
                  prefix,actual_arg_name,hier_name,formal_arg_name);
          fprintf(siva_blif_file,"1 1\n");
        }else {
          fprintf(siva_blif_file,".names %s%s %s%s\n",
                  hier_name,formal_arg_name,prefix,actual_arg_name);
          fprintf(siva_blif_file,"1 1\n");
        }
      }
      fprintf(siva_blif_file,".C Flattening module %s\n",
              sub_module->name);

      siva_write_blif_module(hier_name,sub_module);
      siva_free_string(hier_name);
    }
  }
}  
