/**CFile**********************************************************************

  FileName                   [siva_wit_state.c]

  PackageName                [siva_core]

  Synopsis                   [witness state operation]

  Author                     [Malay K. Ganai]

  Comment                    [Last Modified: 01/12/01]

  Copyright                  [Copyright (C) 2000 University of Texas at Austin.
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License,
  or (at your option) any later version.]

******************************************************************************/

#include "siva.h"

static unsigned siva_wit_state_heap_id;
static int siva_num_wit_state;


/*---------------------------------------------------------------------------*/
void
siva_init_wit_state(void)
{
  siva_wit_state_heap_id = siva_alloc_heap(siva_heap_unsigned_max_cmp);
  siva_wit_state_pos = ALLOC(unsigned, SIVA_SIM_N_WORDS);
  memset(siva_wit_state_pos,0,SIVA_SIM_N_WORDS * 4);
  siva_num_wit_state = 0;
}

/*---------------------------------------------------------------------------*/
void
siva_quit_wit_state(void)
{
  FREE(siva_wit_state_pos);
  siva_release_heap(siva_wit_state_heap_id);
}
  
/*--------------------------------------------------------------------------*/
int 
siva_get_seed_wit_state(SIVA_STATE **state_p,
                        var_set_t *seed_state_vector)
{
  SIVA_KEY key;
  SIVA_CONV conv;
  while(siva_get_wit_state(&key,&conv)){
    /* regenerate state and check if valid state*/
    if (!siva_regenerate_state_vector(conv.state,
                                      seed_state_vector)){
      continue;
    }
    *state_p = conv.state;
    SIVA_VERB1_ARG3("WT: Seed state value = %x, key = %u\n",
                    conv.state->state_value,key.unsigned_key);
    return 1;
  }  
  return 0;
}

/*--------------------------------------------------------------------------*/
void
siva_put_wit_state(unsigned priority_value, SIVA_STATE *state)
{
  SIVA_KEY key;
  SIVA_CONV conv;

  if (priority_value == 0){
    return;
  }

  key.unsigned_key = priority_value;
  conv.state = state;
  siva_heap_insert(key,conv,siva_wit_state_heap_id);
  siva_num_wit_state++;

}

/*--------------------------------------------------------------------------*/
int
siva_get_wit_state(SIVA_KEY *key_ptr, SIVA_CONV *conv_ptr)
{
  int found;

  found = siva_heap_extract_top(key_ptr,conv_ptr,
                                siva_wit_state_heap_id);
  if (found) {
    assert(siva_num_wit_state);
    siva_num_wit_state--;
  }else {
    assert(!siva_num_wit_state);
  }
    
  return found;
}


